/**
 * Copyright (c) 2013, impossibl.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of impossibl.com nor the names of its contributors may
 *    be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package com.impossibl.postgres.jdbc;

import com.impossibl.postgres.api.jdbc.PGAnyType;
import com.impossibl.postgres.api.jdbc.PGConnection;
import com.impossibl.postgres.api.jdbc.PGType;
import com.impossibl.postgres.system.Version;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

@RunWith(JUnit4.class)
public class PGTypeTest {

  private Connection conn;

  @Before
  public void before() throws Exception {
    conn = TestUtil.openDB();
  }

  @After
  public void after() throws SQLException {
    TestUtil.closeDB(conn);
  }

  @Test(expected = SQLException.class)
  public void testVersionCheckThrowsException() throws SQLException {

    PGAnyType invalidType = new PGAnyType() {
      @Override
      public String getName() {
        return "invalid";
      }

      @Override
      public String getVendor() {
        return "Nobody";
      }

      @Override
      public Integer getVendorTypeNumber() {
        return null;
      }

      @Override
      public Version getRequiredVersion() {
        return Version.parse("99");
      }

      @Override
      public Class<?> getJavaType() {
        return Object.class;
      }
    };

    try (PreparedStatement preparedStatement = conn.prepareStatement("SELECT ?::text")) {
      preparedStatement.setObject(1, null, invalidType);
    }

  }

  @Test
  public void testOidsMatch() throws SQLException {

    try (Statement statement = conn.createStatement()) {

      try (ResultSet resultSet = statement.executeQuery("SELECT oid, typname FROM pg_type")) {

        Map<Integer, String> typeOidMap = new HashMap<>();
        while (resultSet.next()) {
          typeOidMap.put(resultSet.getInt(1), resultSet.getString(2));
        }

        for (PGType type : PGType.values()) {
          if (type.getVendorTypeNumber() == null) continue;

          PGConnection conn = (PGConnection) this.conn;
          if (!conn.isServerMinimumVersion(type.getRequiredVersion().getMajor(), type.getRequiredVersion().getMinor())) {
            System.out.println("Skipping test for '" + type.getName() + "' - requires " + type.getRequiredVersion());
            continue;
          }

          String expectedName = typeOidMap.get(type.getVendorTypeNumber());
          assertNotNull("No type with OID="  + type.getVendorTypeNumber() + " found", expectedName);
          assertEquals("Type with OID=" + type.getVendorTypeNumber() + " has incorrect name", expectedName, type.getName());
        }
      }

    }

  }

}
